/*
    Copyright (c) 2010, 2011 aizo ag, Zurich, Switzerland

    Author: Sergey 'Jin' Bostandzhyan <jin@dev.digitalstrom.org>

    This file is part of dS Assistant.

    dS-Assistant is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    dS-Assistant is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

    NOTE: all APIs, especially the JSON API, offered by this application
          are subject to change without further notice.
*/

#ifdef HAVE_CONFIG_H
#include "autoconfig.h"
#endif

#include <limits.h>

#include "json_task_web_handler.h"
#include "ds-assistant.h"
#include "exceptions.h"
#include "database.h"

#define DEFAULT_LOAD_LIMIT  20
#define DEFAULT_LOAD_OFFSET 0

namespace dSA
{


JSONTaskWebHandler::JSONTaskWebHandler(struct mg_connection* connection,
                               const struct mg_request_info* request_info) :
                               JSONWebHandler(connection, request_info) {}

std::string JSONTaskWebHandler::getCurrent()
{
    std::ostringstream result;

    dS_Assistant *dsa = dS_Assistant::getInstance();

    boost::shared_ptr<const dSLongTaskInfo> info = dsa->getCurrentTaskInfo();

    json_object *ret = json_ok();
    json_object *nfo;

    if (info != NULL)
    {
        nfo = task2json(info, true, false);
    }
    else
    {
        nfo = json_object_new_object();
        json_object_object_add(nfo, "id",
            json_object_new_string(util::str(dsa->getInvalidTaskID()).c_str()));
    }

    json_object_object_add(ret, "data", nfo);
    result << json_object_to_json_string(ret);
    json_object_put(ret);

    return result.str();
}

std::string JSONTaskWebHandler::status()
{
    std::ostringstream result;
    dS_Assistant *dsa = dS_Assistant::getInstance();
    boost::shared_ptr<const dSLongTaskInfo> info;
    boost::shared_ptr<std::list<boost::shared_ptr<const dSLongTaskInfo> > > tl =
        dsa->getTaskList();

    std::string task_id = m_parameters["id"];
    if (task_id.empty())
    {
        throw _Exception("missing task id parameter");
    }
    long long int tid = util::str2llint(task_id);

    while (!tl->empty())
    {
        boost::shared_ptr<const dSLongTaskInfo> task = tl->front();
        tl->pop_front();
        if (task->getTaskID() == tid)
        {
            info = task;
            break;
        }
    }

    json_object *ret = json_ok();
    json_object *nfo;

    if (info != NULL)
    {
        nfo = task2json(info, true, true);
    }
    else
    {
        nfo = json_object_new_object();
        json_object_object_add(nfo, "id",
            json_object_new_string(util::str(dsa->getInvalidTaskID()).c_str()));
    }

    json_object_object_add(ret, "data", nfo);
    result << json_object_to_json_string(ret);
    json_object_put(ret);

    return result.str();
}

std::string JSONTaskWebHandler::cancel()
{
    std::ostringstream result;
    dS_Assistant *dsa = dS_Assistant::getInstance();

    std::string task_id = m_parameters["id"];

    if (task_id.empty())
    {
        throw _Exception("missing task id parameter");
    }

    dsa->cancelTask(util::str2llint(task_id));

    json_object *ret = json_ok();
    result << json_object_to_json_string(ret);
    json_object_put(ret);
    return result.str();
}

#ifdef DEBUG
std::string JSONTaskWebHandler::pause()
{
    std::ostringstream result;
    dS_Assistant *dsa = dS_Assistant::getInstance();

    std::string task_id = m_parameters["id"];

    if (task_id.empty())
    {
        throw _Exception("missing task id parameter");
    }

    dsa->pauseTask(util::str2llint(task_id));

    json_object *ret = json_ok();
    result << json_object_to_json_string(ret);
    json_object_put(ret);
    return result.str();
}
#endif

std::string JSONTaskWebHandler::list()
{
    std::ostringstream result;
    dS_Assistant *dsa = dS_Assistant::getInstance();

    boost::shared_ptr<std::list<boost::shared_ptr<const dSLongTaskInfo> > > tl =
        dsa->getTaskList();

    json_object *ret = json_ok();
    json_object *data = json_object_new_object();
    json_object *tasks = json_object_new_array();
    json_object_object_add(data, "tasks", tasks);

    while (!tl->empty())
    {
        boost::shared_ptr<const dSLongTaskInfo> task = tl->front();
        tl->pop_front();

        json_object *nfo = task2json(task);
        json_object_array_add(tasks, nfo);
    }

    json_object_object_add(ret, "data", data);
    result << json_object_to_json_string(ret);
    json_object_put(ret);
    return result.str();
}

std::string JSONTaskWebHandler::load()
{
    std::ostringstream result;
    long long int limit = getOptionalLLIntParam(m_parameters["limit"],
                                                DEFAULT_LOAD_LIMIT,
                                                "limit");
    long long int offset = getOptionalLLIntParam(m_parameters["offset"],
                                                 DEFAULT_LOAD_OFFSET,
                                                 "offset");

    if (limit < 0)
    {
        throw _Exception("invalid value for limit parameter");
    }

    if (offset < 0)
    {
        throw _Exception("invalid value for offset parameter");
    }

    boost::shared_ptr<std::list<boost::shared_ptr<const dSLongTaskInfo> > > tl =
        Database::getInstance()->loadTasks(limit, offset);

    json_object *ret = json_ok();
    json_object *data = json_object_new_object();
    json_object *tasks = json_object_new_array();
    json_object_object_add(data, "tasks", tasks);
    json_object_object_add(ret, "data", data);

    while (!tl->empty())
    {
        boost::shared_ptr<const dSLongTaskInfo> task = tl->front();
        tl->pop_front();

        json_object *nfo = task2json(task);
        json_object_array_add(tasks, nfo);
    }

    result << json_object_to_json_string(ret);
    json_object_put(ret);
    return result.str();
}

std::string JSONTaskWebHandler::resultSummary()
{
    std::ostringstream result;

    std::string task_id = m_parameters["id"];

    if (task_id.empty())
    {
        throw _Exception("missing task id parameter");
    }

    boost::shared_ptr<Database::task_result_summary> summary =
        Database::getInstance()->getTaskResultSummary(util::str2llint(task_id));

    json_object *ret = json_ok();
    json_object_object_add(ret, "data", summary2json(summary));
    result << json_object_to_json_string(ret);
    json_object_put(ret);
    return result.str();
}

json_object* JSONTaskWebHandler::task2json(
                                 boost::shared_ptr<const dSLongTaskInfo> task,
                                 bool progress, bool timestamps)
{
    json_object *t = json_object_new_object();
    json_object_object_add(t, "id",
        json_object_new_string(util::str(task->getTaskID()).c_str()));
    json_object_object_add(t, "name",
        json_object_new_string(task->getName().c_str()));
    json_object_object_add(t, "description",
        json_object_new_string(task->getDescription().c_str()));
    json_object_object_add(t, "status",
        json_object_new_string(task->getTaskStatusText().c_str()));
    json_object_object_add(t, "statusInfo",
        json_object_new_string(task->getStatusInfo().c_str()));
    json_object_object_add(t, "type",
        json_object_new_int((int) task->getTaskType()));

    if (progress)
    {
        json_object_object_add(t, "progress",
            json_object_new_int(task->getProgress()));

        std::string prio = "unknown";
        switch (task->getPriority())
        {
            case High:
                prio = "high";
                break;
            case Normal:
                prio = "normal";
                break;
            default:
                break;
        }

        json_object_object_add(t, "priority",
            json_object_new_string(prio.c_str()));
    }

    if (timestamps)
    {
        json_object_object_add(t, "started",
            json_object_new_string(util::str(task->getStartedTime()).c_str()));
        json_object_object_add(t, "finished",
            json_object_new_string(util::str(task->getFinishedTime()).c_str()));
#ifdef ADVANCED_TASK_MANAGEMENT
        json_object_object_add(t, "queued",
            json_object_new_string(util::str(task->getQueuedTime()).c_str()));
#endif
    }
    return t;
}

json_object* JSONTaskWebHandler::summary2json(
        boost::shared_ptr<Database::task_result_summary> s)
{
    json_object *ret = json_object_new_object();

    json_object *task = task2json(s->task, false, true);
    json_object_object_add(ret, "task", task);

    json_object *params = json_object_new_object();
    json_object_object_add(task, "parameters", params);

    try
    {
        webutil::query task_parameters;
        webutil::parse_query(s->task->getParameters(), task_parameters);
        webutil::query::const_iterator j = task_parameters.begin();
        while (j != task_parameters.end())
        {
            json_object_object_add(params, j->first.c_str(),
                json_object_new_string(j->second.c_str()));
            j++;
        }
    }
    catch (Exception ex)
    {
        fprintf(stderr, "Failed to parse task parameters for task %s\n",
            util::str(s->task->getTaskID()).c_str());
    }

    json_object *res = json_object_new_object();
    json_object_object_add(task, "result", res);

    Database::result_data::const_iterator i = s->data->begin();
    while (i != s->data->end())
    {
        try
        {
            // data type
            switch (i->second->second)
            {
                case SQLITE_INTEGER:
                    {
                        long long int l = util::str2llint(i->second->first);
                        if ((l > INT_MAX) || (l < INT_MIN))
                        {
                            json_object_object_add(res, i->first.c_str(),
                                json_object_new_string(
                                    i->second->first.c_str()));
                        }
                        else
                        {
                            json_object_object_add(res, i->first.c_str(),
                                json_object_new_int(util::str2int(
                                                        i->second->first)));
                        }
                    }
                    break;
                case SQLITE_FLOAT:
                    {
                        std::ostringstream test;
                        test << util::str2double(i->second->first);
                    }
                    json_object_object_add(res, i->first.c_str(),
                        json_object_new_double(util::str2double(
                            i->second->first)));
                    break;
                case SQLITE_TEXT:
                    json_object_object_add(res, i->first.c_str(),
                        json_object_new_string(i->second->first.c_str()));
                    break;
                case SQLITE_NULL:
                case SQLITE_BLOB:
                default:
                    break;
            }
        }
        catch (Exception ex)
        {
            fprintf(stderr, "conversion of %s failed, %s\n",
                    i->first.c_str(), ex.getMessage().c_str());
        }

        i++;
    }

    return ret;
}

void JSONTaskWebHandler::process()
{
    try
    {
        std::string result;
        std::string header = http_ok();

        if (m_command == "getCurrent")
        {
            result = getCurrent();
        }
        else if (m_command == "status")
        {
            result = status();
        }
        else if (m_command == "cancel")
        {
            result = cancel();
        }
#ifdef DEBUG
        else if (m_command == "pause")
        {
            result = pause();
        }
#endif
        else if (m_command == "list")
        {
            result = list();
        }
        else if (m_command == "load")
        {
            result = load();
        }
        else if (m_command == "resultSummary")
        {
            result = resultSummary();
        }
        else
        {
            failure("unknown function called");
            return;
        }

        mg_write(m_connection, header.c_str(), header.length());
        mg_write(m_connection, result.c_str(), result.length());
    }
    catch (Exception ex)
    {
        failure(ex.getMessage().c_str());
    }
    catch (...)
    {
        failure("exception raised while processing call " + m_command);
    }
}

} // namespace

